#/******************************************************************************
#*                              __divsf3
#*******************************************************************************
#* Divide two floating point values: C = A / B
#* NOTE: Denormals are not supported by this function
#*
#* Inputs:  [R1,R2] - 32-bit float, value A
#*          [R3,R4] - 32-bit float, value B
#*
#* Returns: [R1,R2] - 32-bit result, value C
#******************************************************************************/
  def __divsf3
  even
__divsf3:
  # Store return pointer for later
  ai r10, -6
  mov r11, *r10
  mov r1, @2(r10)
  mov r2, @4(r10)

  # If one of the arguments is NaN, return NaN
  bl @__using_nan
  jeq div_return_nan

  # if value B is zero, return NaN
  mov r3, r5
  soc r3, r5
  jne div_calc
div_return_nan:
  li r1, 0x7FC0
  clr r2
  jmp div_end

div_calc:
  # Bits of mantissa used to calculate estimate
  #
  #                      .--------------------.
  #          seee eeee  emmm mmmm  mmmm mmmm  mmmm mmmm
  #          |'---------''----''--------------------'
  # sign     -'  |          |     |
  # exponent ----'          |     |
  # segment  ---------------'     |
  # offset   ---------------------'
  #
  # Use piecewise linearization to approximate 1/B, then use Newton's method
  # to get full precision. If X0 has at least 12 bits or significance, only
  # one round of Newton's method is required.
  #
  #   X0 = linear_estimate(1/B)
  #   1/B = X0(2-B*X0)
  #   1/B = 2*X0 - B*X0^2

  # Calculate exponent for 1/B, store in R0
  #   (0x7f00 - (B >> 16)) & 0xff80
  mov r3, r0
  ai  r0, -0x7f00
  inv r0
  andi r0, 0xff80

  # Find linearization segment
  #   segment = (B >> 18) & 0x1f
  # Convert to byte offset into table
  #   segment_byte = segment * 2
  #                = (B >> 17) & 0x3e
  mov r3, r5   
  srl r5, 1      
  andi r5, 0x3e  # R5 = (B >> 17) & 0x3e
  ai r5, invert_table  # R5 is pointer to highest value of linear segment

  # Find offset into linear segment
  #  offset = (B >> 2) & 0xffff
  mov r4, r8
  srl r8, 2
  mov r3, r7
  sla r7, 14
  soc r7, r8    # R8 is offset into segment

  # Find initial approximation X0, store in [R6,R7]
  mov *r5+, r6  # R6 is base value of linear segment
  mov r6, r7    # Copy base value to calculate slope   
  s   *r5, r7   # R7 is linear segment slope
  mpy r8, r7    # [R7,R8] is (offset*slope)
  inv r7        
  a r6,r7       # R7 is (base-(offset*slope)) >> 16

  # The least significant word of X0 can be ignored as only 12 
  # bits of significance are required for an initial estimate.
  # Save this value for later
  mov r7, r12  # R12 is copy of 16 most significant bts of X0

  # Calculate X0^2, store in [R1,R2]
  mov r7, r1
  mpy r1, r1

  # Shift X0^2 left if needed to set most signficant bit
  mov r1,r1
  jlt div_calc_2
  sla r1, 1
  sla r2, 1
  jnc div_calc_2
  inc r1

  # Shift x0^2 right 8 bits into mantissa position
  # Comments use example of X0 = 0xaabbccdd 
  swpb r2      # X0 = 0xaabbddcc
  swpb r1      # X0 = 0xbbaaddcc
  movb r1, r2  # X0 = 0xbbasbbcc
  sb r1, r1    # X0 = 0x00aabbcc

  # Isolate mantissa of B
div_calc_2:
  ori r3, 0x0080   # Add implied bit
  andi r3, 0x00ff  # Mask exponent bits

  # Calculate B*X0^2
  bl @__mult32

  # See if highest bit is set in result
  jlt div_no_normalize

  # Value needs to be left shifted to set high bit
  sla r1, 1
  sla r2, 1
  jnc $+4
  inc r1
div_no_normalize:

  # Negate B*X0^2
  inv r1
  neg r2
  jnc div_calc_3
  inc r1

div_calc_3:
  # Calculate final mantissa
  # 2*X0 - B*X0^2
  a r12, r1
  a r12, r1

  # Shift right 8 bits into final position
  swpb r2
  swpb r1
  movb r1, r2

  # Add exponent
  andi r1, 0x007f
  soc r0, r1

  # Call multiply to find final result
  mov @2(r10), r3
  mov @4(r10), r4
  bl @__mulsf3

div_end:
  mov *r10, r11
  ai r10, 6
  b *r11

# Look up table of piecewise linearization of 1/(1+(N/32.0)) for N=0 to N=32
# These are the 16 most significant bits of the mantissa
invert_table:
  data 0x0000
  data 0xF83E
  data 0xF0F0
  data 0xEA0E
  data 0xE38E
  data 0xDD67
  data 0xD794
  data 0xD20D
  data 0xCCCC
  data 0xC7CE
  data 0xC30C
  data 0xBE82
  data 0xBA2E
  data 0xB60B
  data 0xB216
  data 0xAE4C
  data 0xAAAA
  data 0xA72F
  data 0xA3D7
  data 0xA0A0
  data 0x9D89
  data 0x9A90
  data 0x97B4
  data 0x94F2
  data 0x9249
  data 0x8FB8
  data 0x8D3D
  data 0x8AD8
  data 0x8888
  data 0x864B
  data 0x8421
  data 0x8208
  data 0x8000

