#/******************************************************************************
#*                              __mulsf3
#*******************************************************************************
#* Multiply two floating point values: C = A * B
#* NOTE: Denormals are not supported by this function
#*
#* Inputs:  [R1,R2] - 32-bit float, value A
#*          [R3,R4] - 32-bit float, value B
#*
#* R+eturns: [R1,R2] - 32-bit result, value C
#******************************************************************************/
  def __mulsf3
  even
__mulsf3:
  # Store return pointer for later
  dect r10
  mov r11, *r10

  # Calculate sign bit
  mov r1, r12
  xor  r3, r12
  andi r12, 0x8000        # Mask all but sign bit
 
  # If one of the arguments is NaN, return NaN
  bl @__using_nan
  jeq mul_ret_nan

  # Isolate exponents
  bl @__remove_exponents # Exponents put into r5 and r6
  c r5,r6
  jgt mul_exp_check

  # Put larger exponent in R5
  mov r5, r0
  mov r6, r5
  mov r0, r6
  
mul_exp_check:
  # inf*0=nan
  ci r5, 0x7f80
  jne mul_exp_not_inf

  # One argument is infinite
  abs r6
  jne mul_exp_not_nan
mul_ret_nan:
  # One argument is infinite, the other is zero, return NAN
  li r1, 0x7fc0
  jmp mul_ret_const

mul_exp_not_nan:
  # Other argument not NAN, result is infinite
  jmp mul_ret_inf

mul_exp_not_inf:
  abs r6
  jne mul_calc
  # One argument is not infinite, the other is zero, result is zero
mul_ret_zero:
  clr r1
mul_ret_const:
  clr r2
  jmp mul_sign

mul_calc:
  # Calculate result exponent
  a r6, r5        # Sum exponents for result exponent
  ai r5, -0x3F80  # Restore exponent bias offset

  # Determine result type based on exponent
  jgt mul_not_infinite
  jnc mul_denormal

  # Result is greater than we can represent
mul_ret_inf:
  li r1, 0x7F80          # Set infinite value
  clr r2
  jmp mul_sign           # Return value

  # Result requires denormal representation
mul_denormal:
  # Denormals not supported, just return zero
  jmp mul_ret_zero

mul_not_infinite:
  # Result is not infinite
  soc r5, r12            # Merge exponent and sign

  bl @__mult32

  # See if highest bit is set in result
  jlt mul_no_normalize

  # Value needs to be left shifted to be in canonical position
  sla r1, 1
  sla r2, 1
  jnc $+4
  inc r1
  jmp mul_finalize

mul_no_normalize:
  # Result required additonal bit. Increment exponent to compensate
  ai r12, 0x0080  

mul_finalize:
  # Result in [R1,R2] needs to be right shifted by 8 bits to be in final position
  swpb r1
  swpb r2
  movb r1, r2

  # Clear exponent bits for merge in next instruction
  andi r1, 0x007F

mul_sign:
  # Place sign in final position
  soc  r12, r1

mul_done:
  # Return value
  mov *r10+, r11
  b *r11

muldiv_mult32:
# Do 32-bit multiply
# This is a 64-bit result, but many of the bytes can be ignored
# espicially since we only need the most significant 25 bits
# 
#     .------.--- We want this part
#     |      | 
# 7 6 5 4 3 2 1 0    bytes   A = Ahi(9bits), Alo(16Bits)
# ____xxxx          Ahi*Bhi (18 bits)
#     __xxxxxx      Ahi*Blo (25 bits)   "_" Known to be zero value
#     __xxxxxx      Alo*Bhi (25 bits)   "x" Unknown value
#         xxxxxxxx  Alo*Blo (32 bits)
#
# Register allocation:
#   output  -> [r1,r2]
#   A value -> [r1,r2]  
#   B value -> [r3,r4] = [Bhi,Blo]
#   A copy  -> [r5,r6] = [Ahi,Alo]
#   Temp    -> [r7,r8]

  mov r1, r5  # \
  mov r2, r6  # / Move A to [r5,r6]
  mov r6, r7  # \
  mpy r4, r7  # / Temp = Alo*Blo
  clr r1      # \
  mov r7, r2  # / Out = (Alo*Blo) >> 16
  mov r6, r7  # \
  mpy r3, r7  # / Temp = Alo*Bhi
  mov r7, r1  # \
  a   r8, r2  # | Out += Alo*Bhi
  jnc $+4     # |
  inc r1      # /
  mov r5, r7  # \
  mpy r4, r7  # / Temp = Ahi*Blo
  a   r7, r1  # \
  a   r8, r2  # | Out += Ahi*Blo
  jnc $+4     # |
  inc r1      # /
  mpy r3, r5  # A = Ahi*Bhi
  a   r6, r1  # Out += (Ahi*Bhi) << 16
  b *r11

